"""
The `optimization` package of `prodsys` provides different optimization algorithms for configuration planning of production systems. 

In configuration planning, the goal is to find a configuration of a production system that maximizes a given objective function. The objective function is a function that maps a configuration to a real number. The configuration is a set of parameters that describe the production system. For example, the configuration of a production system could be the number of machines of a certain type. The objective function could be the output of the production system. The goal is then to find the configuration that maximizes the output. With `prodsys.optimization`, you can use the modelling capacities of  `prodsys.adapters` to specifiy configurations and optimize them according to a scenario and chosen degrees of freedom for optimization.

The following degrees of freedom are supported and can be specified by the scenario data attribute `transformations`:

- Adding or removing production resources
- Adding or removing transport resources
- Adding or removing processes of a production resource
- Changing the position of production resources
- Moving processes between production resources
- Changing the control policy of a production resource
- Changing the control policy of a transport resource
- Changing the routing policy of product types

The following algorithms are supported:

- Evolutionary Algorithms (NGSGA-II)
- Simulated Annealing
- Tabu Search
- Mathematical optimization (Gurobi, only a restricted set of degrees of freedom is supported)

Whilst the mathematical optimization uses a mathematical model for evalutation of production system performance, the other algorithms use `prodsys.simulation` for evaluation. This allows these algorithms to consider more degrees of freedom and optimize according to multiple objectives.

The following KPIs are supported for objectives in optimization with the simulation based optimization algorithms:

- Minimize the WIP (Work in Progress)
- Minimize the throughput time
- Minimize the reconfiguration cost considering capital expenditure (CAPEX)
- Maximize the output / throughput


For a more detailed explanation on the algorithms, degrees of freedom and choice of objectives, please consider the following literature: [Paper](https://doi.org/10.15488/13440)
"""

VERBOSE = 1
"""
The verbosity level of the optimization algorithms. The higher the level, the more information is printed to the console.
"""

from prodsys.optimization.evolutionary_algorithm import evolutionary_algorithm_optimization
from prodsys.optimization.simulated_annealing import simulated_annealing_optimization
from prodsys.optimization.tabu_search import tabu_search_optimization
from prodsys.optimization.math_opt import mathematical_optimization


from __future__ import annotations
from hashlib import md5
from typing import Optional, Union, List, Dict, TYPE_CHECKING
from pydantic import ConfigDict, model_validator
from prodsys.models.core_asset import CoreAsset

if TYPE_CHECKING:
    from prodsys.adapters.adapter import ProductionSystemAdapter

class ProductData(CoreAsset):
    """
    Class that represents product data, specifically the required processes and the allows tranport process.

    The processes describe thereby the process model that needs to be completed for the product to be finished. There are three different ways to describe the process model:

    - Sequential process model: The processes are given as a list of process IDs. The processes are executed sequentially.
    - Adjacency matrix process model: The processes are given as an adjacency matrix that describes the precedence graph of the product.
    - Edges process model: The processes are given as a list of edges. The edges describe the precedence graph of the product.

    See the examples for more insights.

    Args:
        ID (str): ID of the product. If not given, the product type is used. Gets overwritten to the instance product ID, when an instance is created during simulation.
        description (str): Description of the product.
        product_type (str): Type of the product. If not given, the ID is used.
        processes (Union[List[str], List[List[str]], Dict[str, List[str]]]): Processes of the product. This can be a list of process IDs, a list of edges or an adjacency matrix.
        transport_process (str): Transport process of the product.
        aauxiliaries (List[str], optional): List of auxiliary components required to process or transport the product. Defaults to [].

    Examples:
        Product with sequential process model:
        ``` py
        import prodsys
        prodsys.product_data.ProductData(
            ID="Product_1",
            description="product 1",
            product_type="Product_1",
            processes=["P1", "P2", "P3"],
            transport_process="TP1",
        )
        ```

        Product with adjacency matrix process model:
        ``` py
        import prodsys
        prodsys.product_data.ProductData(
            ID="Product_1",
            description="Product 1",
            product_type="Product_1",
            processes={
                "P1": ["P2", "P3"],
                "P2": ["P3"],
                "P3": [],
            },
            transport_process="TP1",
        )
        ```

        Product with graph edges process model:
        ``` py
        import prodsys
        prodsys.product_data.ProductData(
            ID="Product_1",
            description="Product 1",
            product_type="Product_1",
            processes=[
                ["P1", "P2"],
                ["P1", "P3"],
                ["P2", "P4"],
                ["P3", "P4"],
            ],
            transport_process="TP1",
        )
        ```
    """

    product_type: str
    processes: Union[List[str], List[List[str]], Dict[str, List[str]]]
    transport_process: str
    auxiliaries: Optional[List[str]] = []

    def hash(self, adapter: ProductionSystemAdapter) -> str:
        """
        Returns a unique hash of the product considering the processes and the transport process. Can be used to compare products for equal functionality.

        Args:
            adapter (ProductionSystemAdapter): Adapter that contains the process data.

        Raises:
            ValueError: If the process or transport process is not found in the adapter.

        Returns:
            str: Hash of the product.
        """
        processes_hashes = []
        transport_process_hash = ""

        if not isinstance(self.processes, list) or isinstance(self.processes[0], list):
            # TODO: Implement hash for adjacency matrix and edges process models
            raise NotImplementedError("Only list of processes is supported for hashing. Complex process models are not supported yet.")
        
        for process_id in self.processes:
            for process in adapter.process_data:
                if process.ID == process_id:
                    processes_hashes.append(process.hash(adapter))
                    break
            else:
                raise ValueError(f"Process with ID {self.processes} not found for product {self.ID}.")
            
        # TODO: add hashing for auxiliaries!
        
        for transport_process in adapter.process_data:
            if transport_process.ID == self.transport_process:
                transport_process_hash = transport_process.hash(adapter)
                break
        else:
            raise ValueError(f"Transport process with ID {self.transport_process} not found for product {self.ID}.")
        
        return md5("".join([*processes_hashes, transport_process_hash]).encode("utf-8")).hexdigest()
                   
    @model_validator(mode="before")
    def check_processes(cls, values):
        if "product_type" in values and values["product_type"]:
            values["ID"] = values["product_type"]
        else:
            values["product_type"] = values["ID"]
        return values
    
    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "ID": "Product_1",
                "description": "Product with sequential process",
                "product_type": "Product_1",
                "processes": ["P1", "P2", "P3"],
                "transport_process": "TP1",
                "auxiliaries": ["Pallette"]
            },
            {
                "ID": "Product_1",
                "description": "Process with adjacency matrix process",
                "product_type": "Product_1",
                "processes": {
                    "P1": ["P2", "P3"],
                    "P2": ["P3"],
                    "P3": [],
                },
                "transport_process": "TP1",
            },
            {
                "ID": "Product_1",
                "description": "Process with graph edges process",
                "product_type": "Product_1",
                "processes": [
                    ["P1", "P2"],
                    ["P1", "P3"],
                    ["P2", "P4"],
                    ["P3", "P4"],
                ],
                "transport_process": "TP1",
            },
        ]
    })


from __future__ import annotations

from typing import Callable, List, Dict, TYPE_CHECKING
import numpy as np
from numpy.random import exponential, normal, lognormal
from enum import Enum

if TYPE_CHECKING:
    from prodsys.models.time_model_data import FunctionTimeModelData

class FunctionTimeModelEnum(str, Enum):
    Constant = "constant"
    Exponential = "exponential"
    Normal = "normal"
    Lognormal = "lognormal"


def get_constant_list(time_model_data: FunctionTimeModelData) -> List[float]:
    """
    Returns a list of constant values.

    Args:
        time_model_data (FunctionTimeModelData): The time model data.

    Returns:
        List[float]: A list of constant values.
    """
    return [time_model_data.location] * time_model_data.batch_size


def get_exponential_list(time_model_data: FunctionTimeModelData) -> List[float]:
    """
    Returns a list of exponentially distributed values.

    Args:
        time_model_data (FunctionTimeModelData): The time model data.

    Returns:
        List[float]: A list of exponentially distributed values.
    """
    return list(exponential(time_model_data.location, time_model_data.batch_size))


def get_normal_list(time_model_data: FunctionTimeModelData) -> List[float]:
    """
    Returns a list of normally distributed values.

    Args:
        time_model_data (FunctionTimeModelData): The time model data.

    Returns:
        List[float]: A list of normally distributed values.
    """
    return list(normal(time_model_data.location, time_model_data.scale, time_model_data.batch_size))

def get_lognormal_list(time_model_data: FunctionTimeModelData) -> List[float]:
    """
    Returns a list of lognormally distributed values.

    Args:
        time_model_data (FunctionTimeModelData): The time model data.

    Returns:
        List[float]: A list of lognormally distributed values.
    """
    mu = np.log(time_model_data.location **2 / np.sqrt(time_model_data.location**2 + time_model_data.scale**2))
    sigma = np.sqrt(np.log(1 + (time_model_data.scale**2 / time_model_data.location**2)))
    exponential_values = lognormal(mu, sigma, time_model_data.batch_size)
    return list(exponential_values)

FUNCTION_DICT: Dict[str, Callable[[float, float, int], List[float]]] = {
    FunctionTimeModelEnum.Normal: get_normal_list,
    FunctionTimeModelEnum.Constant: get_constant_list,
    FunctionTimeModelEnum.Exponential: get_exponential_list,
    FunctionTimeModelEnum.Lognormal: get_lognormal_list,
}
"""
Dictionary that maps the time model function enum to the corresponding function.
"""



from __future__ import annotations
from collections.abc import Iterable

import random
import os

import numpy as np
from typing import Any, List, Generator
import warnings

warnings.filterwarnings("ignore", category=RuntimeWarning)

from os import listdir
from os.path import isfile, join

import simpy
from prodsys import adapters


def get_class_from_str(name: str, cls_dict: dict):
    """
    Returns the class for a given name from a dictionary containing classes.

    Args:
        name (str): Name of the class.
        cls_dict (dict): Dictionary containing classes.

    Raises:
        ValueError: If the class is not implemented.

    Returns:
        _type_: The class.
    """
    if name not in cls_dict.keys():
        raise ValueError(f"Class '{name}' is not implemented.")
    return cls_dict[name]


def set_seed(seed: int) -> None:
    """
    Sets the seed for numpy and random.

    Args:
        seed (int): The seed.
    """
    np.random.seed(seed)
    random.seed(seed)


def trivial_process(env: simpy.Environment) -> Generator:
    """
    Trivial process that does nothing and is immediately finished.

    Args:
        env (simpy.Environment): The simulation environment.

    Yields:
        Generator: 0 timeout.
    """
    yield env.timeout(0.0)


def read_initial_solutions(
    folder_path: str, base_configuration: adapters.ProductionSystemAdapter
) -> List[adapters.ProductionSystemAdapter]:
    """
    Reads all initial solutions from a folder and returns them as a list of adapters.

    Args:
        folder_path (str): The folder path where the initial solutions are stored.
        base_configuration (adapters.ProductionSystemAdapter): The base configuration for optimization containing the scenario data.

    Returns:
        List[adapters.ProductionSystemAdapter]: List of adapters of the initial solutions.
    """
    file_paths = [f for f in listdir(folder_path) if isfile(join(folder_path, f))]
    adapter_objects = []
    for counter, file_path in enumerate(file_paths):
        if ".json" not in file_path or file_path == "optimization_results.json":
            continue
        adapter = adapters.JsonProductionSystemAdapter()
        adapter.read_data(join(folder_path, file_path))
        adapter.scenario_data = base_configuration.scenario_data.model_copy()
        if not adapter.ID:
            adapter.ID = f"initial_solution_{counter}"
        adapter_objects.append(adapter)
    return adapter_objects

def get_initial_solution_file_pth(
        folder_path: str
) -> str:
    """
    Chooses an initial solution file path from initial solution folder path.

    Args:
        folder_path (str): The folder path where the initial solutions are stored.

    Returns:
        str: The file path of the initial solution.    
    """
    file_paths = [join(folder_path, f) for f in listdir(folder_path) if isfile(join(folder_path, f))]
    return random.choice(file_paths)

def prepare_save_folder(file_paths: str):
    """
    Creates a folder if it does not exist.

    Args:
        file_paths (str): The file path where the folder should be created.
    """
    isExist = os.path.exists(file_paths)
    if not isExist:
        os.makedirs(file_paths)


def flatten(xs: Iterable) -> Iterable:
    """
    Flattens a nested Iterable containing at the lowest level only str or bytes.

    Args:
        xs (Iterable): The nested list.

    Yields:
        Iterable: The flattened list.
    """
    for x in xs:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x

def flatten_object(xs: list) -> Generator[Any, Any, list]:
    """
    Flattens a nested list.

    Args:
        xs (list): The nested list.

    Yields:
        list: The flattened list.
    """
    for x in xs:
        if isinstance(x, list):
            yield from flatten_object(x)
        else:
            yield x


def run_from_ipython() -> bool:
    """
    Checks if the code is run from an ipython notebook.

    Returns:
        bool: True if run from ipython notebook, False otherwise.
    """
    try:
        __IPYTHON__ # type: ignore
        return True
    except NameError:
        return False






from typing import Union, Optional, Generator

from pydantic import BaseModel, ConfigDict, Field

import logging
logger = logging.getLogger(__name__)

import numpy as np
from simpy import events

from prodsys.models import product_data

from prodsys.simulation import (
    request,
    auxiliary,
    process,
    router,
    resources,
    sim,
    sink,
    source,
    proces_models,
    node
)
from prodsys.simulation.state import StateTypeEnum, StateEnum

class ProductInfo(BaseModel):
    """
    Class that represents information of the current state of a product.

    Args:
        resource_ID (str): ID of the resource that the product is currently at.
        state_ID (str): ID of the state that the product is currently at.
        event_time (float): Time of the event.
        activity (state.StateEnum): Activity of the product.
        product_ID (str): ID of the product.
        state_type (state.StateTypeEnum): Type of the state.
    """

    resource_ID: str = Field(init=False, default=None)
    state_ID: str = Field(init=False, default=None)
    event_time: float = Field(init=False, default=None)
    activity: StateEnum = Field(init=False, default=None)
    product_ID: str = Field(init=False, default=None)
    state_type: StateTypeEnum = Field(init=False, default=None)

    model_config=ConfigDict(extra="allow")

    def log_finish_product(
        self,
        resource: Union[resources.Resource, sink.Sink, source.Source],
        _product: "Product",
        event_time: float,
    ):
        """
        Logs the finish of a product.

        Args:
            resource (Union[resources.Resource, sink.Sink, source.Source]): New resource of the product.
            _product (Product): Product that is finished.
            event_time (float): Time of the event.
        """
        self.resource_ID = resource.data.ID
        self.state_ID = resource.data.ID
        self.event_time = event_time
        self.product_ID = _product.product_data.ID
        self.activity = StateEnum.finished_product
        self.state_type = StateTypeEnum.sink

    def log_create_product(
        self,
        resource: Union[resources.Resource, sink.Sink, source.Source],
        _product: "Product",
        event_time: float,
    ) -> None:
        """
        Logs the creation of a product.

        Args:
            resource (Union[resources.Resource, sink.Sink, source.Source]): New resource of the product.
            _product (Product): Product that is created.
            event_time (float): Time of the event.
        """
        self.resource_ID = resource.data.ID
        self.state_ID = resource.data.ID
        self.event_time = event_time
        self.product_ID = _product.product_data.ID
        self.activity = StateEnum.created_product
        self.state_type = StateTypeEnum.source

    def log_start_process(
        self,
        resource: resources.Resource,
        _product: "Product",
        event_time: float,
        state_type: StateTypeEnum,
    ) -> None:
        """
        Logs the start of a process.

        Args:
            resource (resources.Resource): Resource that the product is processed at.
            _product (Product): Product that is processed.
            event_time (float): Time of the event.
            state_type (state.StateTypeEnum): Type of the state.
        """
        self.resource_ID = resource.data.ID
        self.state_ID = resource.data.ID
        self.event_time = event_time
        self.product_ID = _product.product_data.ID
        self.activity = StateEnum.start_state
        self.state_type = state_type

    def log_end_process(
        self,
        resource: resources.Resource,
        _product: "Product",
        event_time: float,
        state_type: StateTypeEnum,
    ) -> None:
        """
        Logs the end of a process.

        Args:
            resource (resources.Resource): Resource that the product is processed at.
            _product (Product): Product that is processed.
            event_time (float): Time of the event.
            state_type (state.StateTypeEnum): Type of the state.
        """
        self.resource_ID = resource.data.ID
        self.state_ID = resource.data.ID
        self.event_time = event_time
        self.product_ID = _product.product_data.ID
        self.activity = StateEnum.end_state
        self.state_type = state_type

Locatable= Union[resources.Resource, node.Node, source.Source, sink.Sink]

class Product(BaseModel):
    """
    Class that represents a product in the discrete event simulation. For easier instantion of the class, use the ProductFactory at prodsys.factories.product_factory.

    Args:
        env (sim.Environment): prodsys simulation environment.
        product_data (product_data.ProductData): Product data that represents the meta information of the simulation product object.
        process_model (proces_models.ProcessModel): Process model that represents the required manufacturing processes and the current state of the product.
        transport_process (process.Process): Transport process that represents the required transport processes.
        product_router (router.Router): Router that is used to route the product object.
    """

    env: sim.Environment
    product_data: product_data.ProductData
    process_model: proces_models.ProcessModel
    transport_process: Union[process.TransportProcess, process.RequiredCapabilityProcess, process.LinkTransportProcess]
    product_router: router.Router

    next_prodution_process: Optional[process.PROCESS_UNION] = Field(default=None, init=False)
    process: events.Process = Field(default=None, init=False)
    current_locatable: Locatable = Field(default=None, init=False)
    finished_process: events.Event = Field(default=None, init=False)
    product_info: ProductInfo = Field(default_factory=ProductInfo, init=False)

    model_config=ConfigDict(arbitrary_types_allowed=True)

    def update_location(self, locatable: Locatable):
        """
        Updates the location of the product object.

        Args:
            locatable (Locatable): Locatable objects where product object currently is.
        """
        self.current_locatable = locatable
        logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "resource": self.current_locatable.data.ID, "event": f"Updated location to {self.current_locatable.data.ID}"})

    def process_product(self):
        self.finished_process = events.Event(self.env)
        self.product_info.log_create_product(
            resource=self.current_locatable, _product=self, event_time=self.env.now
        )
        """
        Processes the product object in a simpy process. The product object is processed after creation until all required production processes are performed and it reaches a sink.
        """
        logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "event": f"Start processing of product"})
        self.set_next_production_process()

        if self.product_data.auxiliaries:
            while True:
                auxiliary_request: request.AuxiliaryRequest = yield self.env.process(self.product_router.route_auxiliary_to_product(self))
                if not auxiliary_request:
                    yield self.env.timeout(0)
                    continue
                break
            logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "resource": auxiliary_request.resource.data.ID, "aux": auxiliary_request.auxiliary.product_data.ID, "process": auxiliary_request.process.process_data.ID, "event": f"auxiliary request for {auxiliary_request.product.product_data.ID}"})
            while True:
                auxiliary_transport_request: request.TransportResquest = yield self.env.process(self.product_router.route_transport_resource_for_item(auxiliary_request))
                if not auxiliary_transport_request:
                    yield self.env.timeout(0)
                    continue
                break
            logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "resource": auxiliary_transport_request.resource.data.ID, "aux": auxiliary_transport_request.product.product_data.ID, "process": auxiliary_transport_request.process.process_data.ID, "origin": auxiliary_transport_request.origin.data.ID, "target": auxiliary_transport_request.target.data.ID, "event": f"starting auxiliary transport request for {auxiliary_transport_request.product.product_data.ID}"})
            yield self.env.process(auxiliary_request.auxiliary.request_process(auxiliary_transport_request))
            logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "resource": auxiliary_transport_request.resource.data.ID, "aux": auxiliary_transport_request.product.product_data.ID, "process": auxiliary_transport_request.process.process_data.ID, "origin": auxiliary_transport_request.origin.data.ID, "target": auxiliary_transport_request.target.data.ID, "event": f"finished waiting for auxiliary transport request for {auxiliary_transport_request.product.product_data.ID}"})

        while self.next_prodution_process:
            logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "process": self.next_prodution_process.process_data.ID, "event": f"Start process of product"})
            while True:
                production_request = yield self.env.process(self.product_router.route_product_to_production_resource(self))
                if not production_request:
                    yield self.env.timeout(0)
                    continue
                break
            while True:
                transport_request = yield self.env.process(self.product_router.route_transport_resource_for_item(production_request))
                if not transport_request:
                    yield self.env.timeout(0)
                    continue
                break
            yield self.env.process(self.request_process(transport_request))
            yield self.env.process(self.request_process(production_request))
            self.set_next_production_process()
        while True:
            transport_to_sink_request = yield self.env.process(self.product_router.route_product_to_sink(self))
            if not transport_to_sink_request:
                yield self.env.timeout(0)
                continue
            break
        yield self.env.process(self.request_process(transport_to_sink_request))
        self.product_info.log_finish_product(
            resource=self.current_locatable, _product=self, event_time=self.env.now
        )
        self.current_locatable.register_finished_product(self)
        logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "event": f"Finished processing of product"})

        if self.product_data.auxiliaries:
            auxiliary_request.auxiliary.update_location(self.current_locatable)
            while True:
                auxiliary_transport_request: request.TransportResquest = yield self.env.process(self.product_router.route_auxiliary_to_store(auxiliary_request.auxiliary))
                if not auxiliary_request:
                    yield self.env.timeout(0)
                    continue
                break
            logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "resource": auxiliary_transport_request.resource.data.ID, "aux": auxiliary_transport_request.product.product_data.ID, "process": auxiliary_transport_request.process.process_data.ID, "origin": auxiliary_transport_request.origin.data.ID, "target": auxiliary_transport_request.target.data.ID, "event": f"starting auxiliary transport request for {auxiliary_transport_request.product.product_data.ID}"})
            yield self.env.process(auxiliary_request.auxiliary.request_process(auxiliary_transport_request))
            auxiliary_request.auxiliary.release_auxiliary_from_product()

    def request_process(self, processing_request: request.Request) -> Generator:
        """
        Requests the next production process of the product object from the next production resource by creating a request event and registering it at the environment.
        """
        if isinstance(processing_request, request.TransportResquest):
            type_ = StateTypeEnum.transport
        else:
            type_ = StateTypeEnum.production
        logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "resource": processing_request.resource.data.ID, "event": f"Request process {processing_request.process.process_data.ID} for {type_}"})
        self.env.request_process_of_resource(
            request=processing_request
        )
        yield self.finished_process
        logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "resource": processing_request.resource.data.ID, "event": f"Finished process {processing_request.process.process_data.ID} for {type_}"})
        self.product_info.log_end_process(
            resource=processing_request.resource,
            _product=self,
            event_time=self.env.now,
            state_type=type_,
        )
        self.finished_process = events.Event(self.env)

    def set_next_production_process(self):
        """
        Sets the next process of the product object based on the current state of the product and its process model.
        """
        next_possible_processes = self.process_model.get_next_possible_processes()
        if not next_possible_processes:
            self.next_prodution_process = None
            logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "event": f"No next process"})
        else:
            self.next_prodution_process = np.random.choice(next_possible_processes)  # type: ignore
            self.process_model.update_marking_from_transition(self.next_prodution_process)  # type: ignore
            logger.debug({"ID": self.product_data.ID, "sim_time": self.env.now, "event": f"Next process {self.next_prodution_process.process_data.ID}"})



from uuid import uuid1
from pydantic.dataclasses import dataclass
from typing import List, Optional, Union
from pydantic import Field
from prodsys.express import core, process
from prodsys.express.queue import Queue
from prodsys.models import auxiliary_data

@dataclass
class Auxiliary(core.ExpressObject):
        """
        Represents an auxiliary object in the production system. The auxiliary can be a tool for a station or a pallet or tray for the transport.
        Auxiliaries are stored at storages, which are represented by the `Storage` class. The auxiliary has a minimum quantity at every storage.
        Processes where the auxiliary is needed, are defined.

        Attributes:
                relevant_processes (List[Union[process.ProductionProcess, process.CapabilityProcess]]): List of relevant production or capability processes.
                relevant_transport_processes (List[process.TransportProcess]): List of relevant transport processes.
                transport_process (process.TransportProcess): The transport process associated with the auxiliary object.
                storages (List[Storage]): List of storages associated with the auxiliary object.
                quantity_in_storages (List[int]): List of initial quantities in the storages.
                ID (Optional[str], optional): The ID of the auxiliary object. Defaults to a generated UUID.
        """

        relevant_processes: List[Union[process.ProductionProcess, process.CapabilityProcess]]
        relevant_transport_processes: List[process.TransportProcess]
        transport_process: process.TransportProcess
        storages: List[Queue]
        quantity_in_storages: List[int]
        ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

        def to_model(self) -> auxiliary_data.AuxiliaryData:
                """
                Converts the Auxiliary object to an AuxiliaryData object.

                Returns:
                        auxiliary_data.AuxiliaryData: The converted AuxiliaryData object.
                """
                return auxiliary_data.AuxiliaryData(
                        ID=self.ID,
                        description="",
                        transport_process=self.transport_process.ID,
                        storages=[storage.ID for storage in self.storages],
                        quantity_in_storages=[initial_quantity_in_store for initial_quantity_in_store in self.quantity_in_storages],
                        relevant_processes=[process.ID for process in self.relevant_processes],
                        relevant_transport_processes=[transportprocess.ID for transportprocess in self.relevant_transport_processes]
                )



        



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, List, Generator, Optional, Union

from pydantic import BaseModel, ConfigDict, Field
import random

import logging
logger = logging.getLogger(__name__)

from simpy.resources import resource
from simpy import events
from prodsys.simulation import sim, store
if TYPE_CHECKING:
    from prodsys.simulation import control, state
    from prodsys.simulation.process import PROCESS_UNION

from prodsys.models.resource_data import (
    RESOURCE_DATA_UNION,
    ProductionResourceData,
    TransportResourceData,
)
from prodsys.util import util

class Resource(BaseModel, ABC, resource.Resource):
    """
    Base class for all resources.

    Args:
        env (sim.Environment): The simpy environment.
        data (RESOURCE_DATA_UNION): The resource data.
        processes (List[PROCESS_UNION]): The processes.
        controller (control.Controller): The controller.
        states (List[state.State]): The states of the resource for breakdowns.
        production_states (List[state.State]): The states of the resource for production.
        setup_states (List[state.SetupState]): The states of the resource for setups.
        got_free (events.Event): The event that is triggered when the resource gets free of processes.
        active (events.Event): The event that is triggered when the resource is active.
        current_setup (PROCESS_UNION): The current setup.
        reserved_setup (PROCESS_UNION): The reserved setup.
    """
    env: sim.Environment
    data: RESOURCE_DATA_UNION
    processes: List[PROCESS_UNION]
    controller: control.Controller

    states: List[state.State] = Field(default_factory=list, init=False)
    production_states: List[state.State] = Field(default_factory=list, init=False)
    setup_states: List[state.SetupState] = Field(default_factory=list, init=False)
    charging_states: List[state.ChargingState] = Field(default_factory=list, init=False)

    got_free: events.Event = Field(default=None, init=False)
    active: events.Event = Field(default=None, init=False)
    current_setup: PROCESS_UNION = Field(default=None, init=False)
    reserved_setup: PROCESS_UNION = Field(default=None, init=False)

    model_config=ConfigDict(arbitrary_types_allowed=True, extra="allow")

    @property
    def capacity_current_setup(self) -> int:
        """
        Returns the capacity of the resource for the current setup with considering that the resource could be in a setup process.

        Returns:
            int: The capacity of the resource for the current setup.
        """
        if not self.current_setup and not self.reserved_setup:
            return self.capacity
        elif (
            self.reserved_setup
            and self.current_setup.process_data.ID
            != self.reserved_setup.process_data.ID
        ):
            current_setup_ID = self.reserved_setup.process_data.ID
        elif self.current_setup:
            current_setup_ID = self.current_setup.process_data.ID
        length = len(
            [
                state
                for state in self.production_states
                if state.state_data.ID == current_setup_ID
            ]
        )
        return length

    def reserve_setup(self, process: PROCESS_UNION) -> None:
        """
        Reserves the setup of the resource for a process. This is used to prevent that capacity is wrong estimated during setup.

        Args:
            process (PROCESS_UNION): The process that wants to reserve the setup.
        """
        self.reserved_setup = process

    def unreserve_setup(self) -> None:
        """
        Unreserves the setup of the resource. This is used to prevent that the resource is used for another process while it is in a setup process.
        """
        self.reserved_setup = None

    @property
    def in_setup(self) -> bool:
        """
        Returns if the resource is in a setup process.

        Returns:
            bool: True if the resource is in a setup process, False otherwise.
        """
        return self.reserved_setup is not None

    @property
    def full(self) -> bool:
        """
        Returns if the resource is full.

        Returns:
            bool: True if the resource is full or in setup, False otherwise.
        """
        if self.in_setup:
            return True
        return (
            self.capacity_current_setup
            - (len(self.controller.running_processes) + self.controller.reserved_requests_count)
        ) <= 0
    
    @property
    def requires_charging(self) -> bool:
        """
        Returns if the resource requires charging.

        Returns:
            bool: True if the resource requires charging, False otherwise.
        """
        return any([state_instance.requires_charging() for state_instance in self.charging_states if isinstance(state_instance, state.ChargingState)])

    
    def charge(self) -> Generator:
        """
        Charges the resource.

        Yields:
            Generator: The type of the yield depends on the resource.
        """
        # TODO: transport AGV to charging station
        for input_state in self.charging_states:
            if not input_state.requires_charging():
                continue
            input_state.prepare_for_run()
            yield self.env.process(input_state.process_state())

    def consider_battery_usage(self, amount: float) -> None:
        """
        Reduces the battery level of the resource.

        Args:
            amount (float): The amount to reduce the battery level.
        """
        for state_instance in self.charging_states:
            state_instance.add_battery_usage_time(amount)
    
    def get_controller(self) -> control.Controller:
        """
        Returns the controller of the resource.

        Returns:
            control.Controller: The controller of the resource.
        """
        return self.controller

    def add_state(self, input_state: state.STATE_UNION) -> None:
        """
        Adds a state to the resource.

        Args:
            input_state (state.STATE_UNION): The state to add.
        """
        if isinstance(input_state, state.SetupState):
            self.setup_states.append(input_state)
        elif isinstance(input_state, state.ChargingState):
            self.charging_states.append(input_state)
        else:
            self.states.append(input_state)
        input_state.set_resource(self)

    def add_production_state(self, input_state: state.ProductionState) -> None:
        """
        Adds a production state to the resource.

        Args:
            input_state (state.ProductionState): The production state to add.
        """
        self.production_states.append(input_state)
        input_state.set_resource(self)

    def start_states(self):
        """
        Starts the simpy processes of the states of the resource in simpy.
        """
        resource.Resource.__init__(self, self.env, capacity=self.data.capacity)
        self.active = events.Event(self.env).succeed()
        self.got_free = events.Event(self.env)
        for actual_state in self.states + self.production_states + self.setup_states + self.charging_states:
            actual_state.activate_state()
        for actual_state in self.states:
            actual_state.process = self.env.process(actual_state.process_state())

    def get_process(self, process: PROCESS_UNION) -> state.State:
        """
        Returns the ProducitonState or CapabilityState of the resource for a process.

        Args:
            process (PROCESS_UNION): The process to get the state for.

        Raises:
            ValueError: If the process is not found in the resource.

        Returns:
            state.State: The state of the resource for the process.
        """
        possible_states = [
            actual_state
            for actual_state in self.production_states
            if actual_state.state_data.ID == process.process_data.ID
        ]
        if not possible_states:
            raise ValueError(
                f"Process {process.process_data.ID} not found in resource {self.data.ID}"
            )
        return random.choice(possible_states)

    def get_processes(self, process: PROCESS_UNION) -> List[state.State]:
        """
        Returns the ProducitonState or CapabilityState of the resource for a process.

        Args:
            process (PROCESS_UNION): The process to get the state for.

        Raises:
            ValueError: If the process is not found in the resource.

        Returns:
            List[state.State]: The state of the resource for the process.
        """
        possible_states = [
            actual_state
            for actual_state in self.production_states
            if actual_state.state_data.ID == process.process_data.ID
        ]
        if not possible_states:
            raise ValueError(
                f"Process {process.process_data.ID} not found in resource {self.data.ID}"
            )
        return possible_states

    def get_free_process(self, process: PROCESS_UNION) -> Optional[state.State]:
        """
        Returns a free ProductionState or CapabilityState of the resource for a process.

        Args:
            process (PROCESS_UNION): The process to get the state for.

        Returns:
            Optional[state.State]: The state of the resource for the process.
        """
        for actual_state in self.production_states:
            if actual_state.state_data.ID == process.process_data.ID and (
                actual_state.process is None or not actual_state.process.is_alive
            ):
                return actual_state
        return None

    def get_location(self) -> List[float]:
        """
        Returns the location of the resource.

        Returns:
            List[float]: The location of the resource. Has to have length 2.
        """
        return self.data.location
    

    def get_input_queue_length(self) -> int:
        """
        Returns total number of items in all input_queues.

        Returns:
            int: Sum of items in the resources input-queues.
        """
        return sum([len(q.items) for q in self.input_queues])

    def get_output_queue_length(self) -> int:
        """
        Returns total number of items in all output_queues.
        
        Returns:
            int: Sum of items in the resources output-queues.
        """
        return sum([len(q.items) for q in self.output_queues])

    def set_location(self, new_location: List[float]) -> None:
        """
        Sets the location of the resource.

        Args:
            new_location (List[float]): The new location of the resource. Has to have length 2.
        """
        self.data.location = new_location

    def get_states(self) -> List[state.State]:
        """
        Returns the states of the resource.

        Returns:
            List[state.State]: The states of the resource.
        """
        return self.states

    def activate(self):
        """
        Activates the resource after a breakdwon.
        """
        if any([state_instance.active_breakdown for state_instance in self.states if isinstance(state_instance, state.BreakDownState)]):
            logger.debug({"ID": self.data.ID, "sim_time": self.env.now, "resource": self.data.ID, "event": f"Breakdown still active that blocks activation of resource"})
            return
        self.active.succeed()

    def interrupt_states(self):
        """
        Interrupts the states of the resource.
        """
        logger.debug({"ID": self.data.ID, "sim_time": self.env.now, "resource": self.data.ID, "event": f"Start interrupting processes of resource"})
        if self.active.triggered:
            self.active = events.Event(self.env)
        for state_instance in self.setup_states + self.production_states:
            if state_instance.process and state_instance.process.is_alive and not state_instance.interrupted:
                state_instance.interrupt_process()
        logger.debug({"ID": self.data.ID, "sim_time": self.env.now, "resource": self.data.ID, "event": f"Interrupted processes of resource"})


    def get_free_of_setups(self) -> Generator:
        """
        Returns a generator that yields when all setups are finished.

        Yields:
            Generator: The generator of the yield, which is yielded when all setups are finished.
        """
        running_setups = [
            state.process
            for state in self.setup_states
            if (state.process and state.process.is_alive)
        ]
        logger.debug({"ID": self.data.ID, "sim_time": self.env.now, "resource": self.data.ID, "event": f"Start waiting for free of setups"})
        yield events.AllOf(self.env, running_setups)
        logger.debug({"ID": self.data.ID, "sim_time": self.env.now, "resource": self.data.ID, "event": f"Finished waiting for free of setups"})

    def get_free_of_processes_in_preparation(self) -> Generator:
        """
        Returns a generator that yields when all processes in preparation are finished.

        Yields:
            Generator: The generator of the yield, which is yielded when all processes in preparation are finished.
        """
        running_processes = [
            state.process
            for state in self.production_states
            if (state.process and state.process.is_alive)
        ]
        logger.debug({"ID": self.data.ID, "sim_time": self.env.now, "resource": self.data.ID, "event": f"Start waiting for free of processes in preparation"})
        yield events.AllOf(self.env, running_processes)
        logger.debug({"ID": self.data.ID, "sim_time": self.env.now, "resource": self.data.ID, "event": f"Finished waiting for free of processes in preparation"})

    def setup(self, _process: PROCESS_UNION) -> Generator:
        """
        Sets up the resource for a process.

        Args:
            _process (PROCESS_UNION): The process to set up the resource for.

        Yields:
            Generator: The type of the yield depends on the process.
        """
        if self.current_setup is None:
            yield self.env.process(util.trivial_process(self.env))
            self.current_setup = _process
            return
        if self.reserved_setup:
            setup_to_compare = self.reserved_setup
        else:
            setup_to_compare = self.current_setup

        if setup_to_compare.process_data.ID == _process.process_data.ID:
            yield self.env.process(self.get_free_of_setups())
            yield self.env.process(util.trivial_process(self.env))
            return

        for input_state in self.setup_states:
            if (
                input_state.state_data.target_setup == _process.process_data.ID
                and input_state.state_data.origin_setup
                == setup_to_compare.process_data.ID
            ):
                self.reserve_setup(_process)
                yield self.env.process(self.get_free_of_setups())
                input_state.prepare_for_run()
                input_state.process = self.env.process(input_state.process_state())
                logger.debug({"ID": self.data.ID, "sim_time": self.env.now, "resource": self.data.ID, "process": _process.process_data.ID, "event": f"Start setup process"})
                yield input_state.process
                logger.debug({"ID": self.data.ID, "sim_time": self.env.now, "resource": self.data.ID, "process": _process.process_data.ID, "event": f"Finished setup process"})
                input_state.process = None
                self.current_setup = _process
                self.unreserve_setup()

        else:
            yield self.env.process(self.get_free_of_setups())
            yield self.env.process(util.trivial_process(self.env))


class ProductionResource(Resource):
    """
    A production resource to perform production processes. Has additionally to a Resource input and output queues and a fixed location.

    Args:
        env (sim.Environment): The simpy environment.
        data (ProductionResourceData): The resource data.
        processes (List[PROCESS_UNION]): The processes.
        controller (control.ProductionController): The controller.
        states (List[state.State]): The states of the resource for breakdowns.
        production_states (List[state.State]): The states of the resource for production.
        setup_states (List[state.SetupState]): The states of the resource for setups.
        got_free (events.Event): The event that is triggered when the resource gets free of processes.
        active (events.Event): The event that is triggered when the resource is active.
        current_setup (PROCESS_UNION): The current setup.
        reserved_setup (PROCESS_UNION): The reserved setup.
        input_queues (List[store.Queue]): The input queues.
        output_queues (List[store.Queue]): The output queues.


    """
    data: ProductionResourceData
    controller: control.ProductionController

    input_queues: List[store.Queue] = []
    output_queues: List[store.Queue] = []

    def add_input_queues(self, input_queues: List[store.Queue]):
        self.input_queues.extend(input_queues)

    def add_output_queues(self, output_queues: List[store.Queue]):
        self.output_queues.extend(output_queues)

    def reserve_input_queues(self):
        for input_queue in self.input_queues:
            input_queue.reserve()

    def unreserve_input_queues(self):
        for input_queue in self.input_queues:
            input_queue.unreseve()

class TransportResource(Resource):
    """
    A transport resource to perform transport processes. Can change its and the product's location during transport processes.

    Args:
        env (sim.Environment): The simpy environment.
        data (TransportResourceData): The resource data.
        processes (List[PROCESS_UNION]): The processes.
        controller (control.TransportController): The controller.
        states (List[state.State]): The states of the resource for breakdowns.
        production_states (List[state.State]): The states of the resource for production.
        setup_states (List[state.SetupState]): The states of the resource for setups.
        got_free (events.Event): The event that is triggered when the resource gets free of processes.
        active (events.Event): The event that is triggered when the resource is active.
        current_setup (PROCESS_UNION): The current setup.
        reserved_setup (PROCESS_UNION): The reserved setup.
    """
    data: TransportResourceData
    controller: control.TransportController


RESOURCE_UNION = Union[ProductionResource, TransportResource]
""" Union Type for Resources. """

from prodsys.simulation import control, state
from prodsys.simulation.process import PROCESS_UNION



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional, Union, TYPE_CHECKING, Generator, List

import logging
logger = logging.getLogger(__name__)

from simpy import events
from simpy import exceptions
from pydantic import BaseModel, ConfigDict, model_validator, Field

from prodsys.simulation import sim, time_model
from prodsys.models.state_data import (
    StateData,
    BreakDownStateData,
    ProductionStateData,
    TransportStateData,
    SetupStateData,
    ProcessBreakDownStateData,
    ChargingStateData
)

if TYPE_CHECKING:
    from prodsys.simulation import product, resources, auxiliary


class StateEnum(str, Enum):
    """
    Enum for the different types a state can be in.
    """
    start_state = "start state"
    start_interrupt = "start interrupt"
    end_interrupt = "end interrupt"
    end_state = "end state"

    created_product = "created product"
    started_product_processing = "started product processing"
    # TODO: maybe rename finished_product to finished_product_processing for consistency
    finished_product = "finished product"

    created_auxiliary = "created auxiliary"
    started_auxiliary_usage = "started auxiliary usage"
    finished_auxiliary_usage = "finished auxiliary usage"


class StateTypeEnum(str, Enum):
    """
    Enum for the different types of states.
    """
    production = "Production"
    transport = "Transport"
    breakdown = "Breakdown"
    process_breakdown = "ProcessBreakdown"
    setup = "Setup"
    source = "Source"
    sink = "Sink"
    store = "Store"
    charging = "Charging"


class StateInfo(BaseModel):
    """
    Class that represents the current event information of a state while simulating.

    Args:
        ID (str): The ID of the state.
        resource_ID (str): The ID of the resource the state belongs to.
        _event_time (Optional[float], optional): The time of the event. Defaults to 0.0.
        _expected_end_time (Optional[float], optional): The expected end time of the state. Defaults to 0.0.
        _activity (Optional[StateEnum], optional): The activity of the state. Defaults to None.
        _state_type (Optional[StateTypeEnum], optional): The type of the state. Defaults to None.
        _product_ID (str, optional): The ID of the product the state belongs to. Defaults to "".
        _target_ID (str, optional): The ID of the target the state belongs to. Defaults to "".
    """
    ID: str
    resource_ID: str
    _event_time: Optional[float] = 0.0
    _expected_end_time: Optional[float] = 0.0
    _activity: Optional[StateEnum] = None
    _state_type: Optional[StateTypeEnum] = None
    _product_ID: str = ""
    _target_ID: str = ""
    _origin_ID: str = ""
    _empty_transport: Optional[bool] = None

    model_config=ConfigDict(extra="allow")

    def log_transport(self, origin: Optional[product.Locatable], target: product.Locatable, state_type: StateTypeEnum, empty_transport: bool):
        """
        Logs the target location of a transport state.

        Args:
            origin (Optional[product.Locatable]): The origin location, either a resource, source, node or a sink.
            target (product.Locatable): The target location, either a resource, source, node or a sink.
            state_type (StateTypeEnum): The type of the state.
            empty_transport (bool): Indicates if the transport is empty.
        """
        if not origin:
            self._origin_ID = "Loading station"
        else:
            self._origin_ID = origin.data.ID
        self._target_ID = target.data.ID
        self._state_type = state_type
        self._empty_transport = empty_transport

    def log_product(self, _product: product.Product, state_type: StateTypeEnum):
        """
        Logs the product of a transport or production state.

        Args:
            _product (product.Product): The product.
            state_type (StateTypeEnum): The type of the state.
        """
        self._product_ID = _product.product_data.ID
        self._state_type = state_type

    def log_auxiliary(self, _auxiliary: auxiliary.Auxiliary, state_type: StateTypeEnum):
        """
        Logs the product of a transport or production state.

        Args:
            _product (product.Product): The product.
            state_type (StateTypeEnum): The type of the state.
        """
        self._product_ID = _auxiliary.data.ID
        self._state_type = state_type

    def log_start_state(
        self, start_time: float, expected_end_time: float, state_type: StateTypeEnum
    ):
        """
        Logs the start of a state.

        Args:
            start_time (float): The start time of the state.
            expected_end_time (float): The expected end time of the state.
            state_type (StateTypeEnum): The type of the state.
        """
        self._event_time = start_time
        self._expected_end_time = expected_end_time
        self._activity = StateEnum.start_state
        self._state_type = state_type

    def log_start_interrupt_state(self, start_time: float, state_type: StateTypeEnum):
        """
        Logs the start of an interrupt of a state.

        Args:
            start_time (float): The start time of the interruption.
            state_type (StateTypeEnum): The type of the state.
        """
        self._event_time = start_time
        self._activity = StateEnum.start_interrupt
        self._state_type = state_type

    def log_end_interrupt_state(
        self, end_time: float, expected_end_time: float, state_type: StateTypeEnum
    ):
        """
        Logs the end of an interrupt of a state.

        Args:
            end_time (float): The end time of the interruption.
            expected_end_time (float): The expected end time of the state.
            state_type (StateTypeEnum): The type of the state.
        """
        self._event_time = end_time
        self._expected_end_time = expected_end_time
        self._activity = StateEnum.end_interrupt
        self._state_type = state_type

    def log_end_state(self, end_time: float, state_type: StateTypeEnum):
        """
        Logs the end of a state.

        Args:
            end_time (float): The end time of the state.
            state_type (StateTypeEnum): The type of the state.
        """
        self._event_time = end_time
        self._activity = StateEnum.end_state
        self._state_type = state_type


def debug_logging(state_instance: State, msg: str):
    """
    Logs a debug message for a state.

    Args:
        state_instance (State): The state.
        msg (str): The message.
    """
    logger.debug({"ID": state_instance.state_data.ID, "sim_time": state_instance.env.now, "resource": state_instance.resource.data.ID, "event": msg})


class State(ABC, BaseModel):
    """
    Abstract class that represents a state of a resource in the simulation. A state has a process that is simulated when the resource starts a state. States can exist in parallel and can interrupt each other.

    Args:
        state_data (StateData): The data of the state.
        time_model (time_model.TimeModel): The time model of the state.
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indidcates if the state is active. Defaults to None.
        finished_process (events.Event, optional): Event that indicates if the state is finished. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
    """
    state_data: StateData
    time_model: time_model.TimeModel
    env: sim.Environment
    active: events.Event = Field(default=None, init=False)
    finished_process: events.Event = Field(default=None, init=False)
    resource: resources.Resource = Field(init=False, default=None)
    process: Optional[events.Process] = Field(default=None)
    state_info: StateInfo = Field(None)

    model_config=ConfigDict(arbitrary_types_allowed=True)


    def set_resource(self, resource_model: resources.Resource) -> None:
        """
        Sets the resource of the state.

        Args:
            resource_model (resources.Resource): The resource the state belongs to.
        """
        self.resource = resource_model
        self.state_info = StateInfo(
            ID=self.state_data.ID, resource_ID=self.resource.data.ID
        )

    def deactivate(self):
        """
        Deactivates the state by setting the active event to a new event which is not yet triggered.
        """
        self.active = events.Event(self.env)

    def activate(self):
        """
        Activates the state by triggering the active event.

        Raises:
            RuntimeError: If the state is allready active.
        """
        try:
            self.active.succeed()
        except:
            raise RuntimeError(f"state {self.state_data.ID} is allready succeded!!")

    @abstractmethod
    def process_state(self) -> Generator:
        """
        Runs a single process of the state. The process is the key component for the behavior of the state while simulating. All the logic of the state is implemented in the process.

        Yields:
            Generator: The generator of the process.
        """
        pass

    @abstractmethod
    def interrupt_process(self):
        """
        Interrupts the process of the state. 
        """
        pass

    def activate_state(self):
        """
        Activates the state and at start of the simulation.
        """
        pass

    def prepare_for_run(self):
        """
        Prepares the state for running the process of a state.
        """
        pass


class ProductionState(State):
    """
    Represents a production state of a resource in the simulation. A production state has a process that simulates the production process which takes some time. The production state continues the creation process of a product. If a resource has a higher capacity than 1 for a process, multiple production states exist, that can run in parallel.

    Args:
        state_data (ProductionStateData): The data of the state.
        time_model (time_model.TimeModel): The time model of the state.
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indidcates if the state is active. Defaults to None.
        finished_process (events.Event, optional): Event that indicates if the state is finished. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
        start (float, optional): The start time of the state. Defaults to 0.0.
        done_in (float, optional): The ramaining time for the state to finish. Defaults to 0.0.
        interrupted (bool, optional): Indicates if the state is interrupted. Defaults to False.
    """
    state_data: ProductionStateData
    start: float = 0.0
    done_in: float = 0.0
    interrupted: bool = False

    def prepare_for_run(self):
        self.finished_process = events.Event(self.env)

    def activate_state(self):
        self.active = events.Event(self.env).succeed()

    def process_state(self) -> Generator:
        self.done_in = self.time_model.get_next_time()
        self.resource.consider_battery_usage(self.done_in)
        while True:
            try:
                if self.interrupted:
                    debug_logging(self, f"interrupted while waiting for activation or activation of resource")
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    self.interrupted = False
                debug_logging(self, f"wait for activation or activation of resource before process")
                yield events.AllOf(self.env, [self.resource.active, self.active])
                break
            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(f"Simpy interrupt occured at {self.state_data.ID} although process is not interrupted")
        while self.done_in:
            try:
                if self.interrupted:
                    self.state_info.log_start_interrupt_state(
                        self.env.now, StateTypeEnum.production
                    )
                    self.update_done_in()
                    debug_logging(self, f"interrupted process that ends in {self.done_in}")
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    debug_logging(self, f"interrupt over for process that ends in {self.done_in}")
                    self.interrupted = False
                    self.state_info.log_end_interrupt_state(
                        self.env.now, self.env.now + self.done_in, StateTypeEnum.production
                    )
                self.start = self.env.now
                self.state_info.log_start_state(
                        self.start, self.start + self.done_in, StateTypeEnum.production
                )
                debug_logging(self, f"starting process that ends in {self.done_in}")
                yield self.env.timeout(self.done_in)
                self.done_in = 0  # Set to 0 to exit while loop.

            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(f"Simpy interrupt occured at {self.state_data.ID} although process is not interrupted")
        debug_logging(self, f"process finished")
        self.state_info.log_end_state(self.env.now, StateTypeEnum.production)
        self.finished_process.succeed()

    def update_done_in(self):
        if self.start == 0:
            return
        self.done_in -= (self.env.now - self.start)  # How much time left?
        if self.done_in < 0:
            self.done_in = 0

    def interrupt_process(self):
        if self.process and self.process.is_alive and not self.interrupted:
            self.interrupted = True
            self.process.interrupt()


class TransportState(State):
    """
    Represents a transport state of a resource in the simulation. A transport state has a process that simulates the transport of a product. The transport state continues the transport process of a product. If a resource has a higher capacity than 1 for a process, multiple transport states exist, that can run in parallel but only with the same target and end location.

    Args:
        state_data (TransportStateData): The data of the state.
        time_model (time_model.TimeModel): The time model of the state.
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indidcates if the state is active. Defaults to None.
        finished_process (events.Event, optional): Event that indicates if the state is finished. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
        start (float, optional): The start time of the state. Defaults to 0.0.
        done_in (float, optional): The ramaining time for the state to finish. Defaults to 0.0.
        interrupted (bool, optional): Indicates if the state is interrupted. Defaults to False.
    """
    state_data: TransportStateData
    start: float = 0.0
    done_in: float = 0.0
    interrupted: bool = False

    def prepare_for_run(self):
        self.finished_process = events.Event(self.env)

    def activate_state(self):
        self.active = events.Event(self.env).succeed()

    def process_state(self, target: List[float], initial_transport_step: bool, last_transport_step: bool) -> Generator:
        self.done_in = self.time_model.get_next_time(
            origin=self.resource.get_location(), target=target
        )
        if initial_transport_step and hasattr(self.time_model, "reaction_time") and self.time_model.time_model_data.reaction_time:
            self.done_in -= self.time_model.time_model_data.reaction_time
        self.resource.consider_battery_usage(self.done_in)

        while True:
            try:
                if self.interrupted:
                    debug_logging(self, f"interrupted while waiting for activation or activation of resource")
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    self.interrupted = False
                debug_logging(self, f"wait for activation or activation of resource before process")
                yield events.AllOf(self.env, [self.resource.active, self.active])
                break
            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(f"Simpy interrupt occured at {self.state_data.ID} although process is not interrupted")
        while self.done_in:
            try:
                if self.interrupted:
                    self.state_info.log_start_interrupt_state(
                        self.env.now, StateTypeEnum.transport
                    )
                    self.update_done_in()
                    debug_logging(self, f"interrupted process that ends in {self.done_in}")
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    debug_logging(self, f"interrupt over for process that ends in {self.done_in}")
                    self.interrupted = False
                    self.state_info.log_end_interrupt_state(
                        self.env.now, self.env.now + self.done_in, StateTypeEnum.transport
                    )
                self.start = self.env.now
                self.state_info.log_start_state(
                self.start, self.start + self.done_in, StateTypeEnum.transport
                    )
                debug_logging(self, f"starting process that ends in {self.done_in}")
                yield self.env.timeout(self.done_in)
                self.done_in = 0  # Set to 0 to exit while loop.
            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(f"Simpy interrupt occured at {self.state_data.ID} although process is not interrupted")
        debug_logging(self, f"process finished")
        self.state_info.log_end_state(self.env.now, StateTypeEnum.transport)
        self.finished_process.succeed()

    def update_done_in(self):
        self.done_in -= (self.env.now - self.start)  # How much time left?
        if self.done_in < 0:
            self.done_in = 0

    def interrupt_process(self):
        if self.process and self.process.is_alive and not self.interrupted:
            self.interrupted = True
            self.process.interrupt()


class BreakDownState(State):
    """
    Represents a breakdown state of a resource in the simulation. A breakdown state has a process that simulates the breakdown of a resource. All other running production, transport or setup states get interrupted.

    Args:
        state_data (BreakDownStateData): The data of the state.
        time_model (time_model.TimeModel): The time model of the state.
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indidcates if the state is active. Defaults to None.
        finished_process (events.Event, optional): Event that indicates if the state is finished. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
        repair_time_model (time_model.TimeModel, optional): The time model of the repair time. Defaults to None.
    """
    state_data: BreakDownStateData
    repair_time_model: time_model.TimeModel
    active_breakdown: bool = False

    @model_validator(mode="before")
    def post_init(cls, values):
        values["active"] = events.Event(values["env"])
        return values

    def process_state(self) -> Generator:
        while True:
            yield self.env.process(self.wait_for_breakdown())
            self.active_breakdown = True
            debug_logging(self, f"breakdown occured, start interrupting processes")
            self.resource.interrupt_states()
            repair_time = self.repair_time_model.get_next_time()
            debug_logging(self, f"interrupted states, starting breakdown for {repair_time}")
            self.state_info.log_start_state(
                self.env.now, self.env.now + repair_time, StateTypeEnum.breakdown
            )
            yield self.env.timeout(repair_time)
            self.active_breakdown = False
            self.resource.activate()
            debug_logging(self, f"breakdown finished, reactivating resource")
            self.state_info.log_end_state(self.env.now, StateTypeEnum.breakdown)

    def wait_for_breakdown(self):
        yield self.env.timeout(self.time_model.get_next_time())

    def interrupt_process(self):
        pass


class ProcessBreakDownState(State):
    """
    Represents a process breakdown state of a resource in the simulation. A process breakdown state has a process that simulates the breakdown of a process of a resource. Only production states of this type of process get interrupted. Also all setup states get interrupted.

    Args:
        state_data (ProcessBreakDownStateData): The data of the state.
        time_model (time_model.TimeModel): The time model of the state.
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indidcates if the state is active. Defaults to None.
        finished_process (events.Event, optional): Event that indicates if the state is finished. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
        production_states (List[State], optional): The production states of the process. Defaults to None.
        repair_time_model (time_model.TimeModel, optional): The time model of the repair time. Defaults to None.
    """
    state_data: ProcessBreakDownStateData
    production_states: List[State] = None
    repair_time_model: time_model.TimeModel

    @model_validator(mode="before")
    def post_init(cls, values):
        values["active"] = events.Event(values["env"])
        return values

    def set_production_states(self, production_states: List[ProductionState]):
        if any(
            [
                production_state.state_data.ID != self.state_data.process_id
                for production_state in production_states
            ]
        ):
            raise ValueError(
                f"Production states {production_states} do not match process id {self.state_data.process_id}"
            )
        self.production_states = production_states

    def process_state(self) -> Generator:
        while True:
            yield self.env.process(self.wait_for_breakdown())
            debug_logging(self, f"breakdown occured, start interrupting processes")
            yield events.AllOf(
                self.env, [state.active for state in self.production_states]
            )
            debug_logging(self, f"interrupted states, starting breakdown for {self.repair_time_model.get_next_time()}")
            for state in self.production_states + self.resource.setup_states:
                state.deactivate()
            for state in self.production_states + self.resource.setup_states:
                if state.process and state.process.is_alive and not state.interrupted:
                    state.interrupt_process()
            repair_time = self.repair_time_model.get_next_time()
            debug_logging(self, f"interrupted states, starting breakdown for {repair_time}")
            self.state_info.log_start_state(
                self.env.now, self.env.now + repair_time, StateTypeEnum.process_breakdown
            )
            yield self.env.timeout(repair_time)
            self.state_info.log_end_state(self.env.now, StateTypeEnum.process_breakdown)
            debug_logging(self, f"breakdown finished, reactivating states")
            for state in self.production_states + self.resource.setup_states:
                state.activate()

    def wait_for_breakdown(self):
        yield self.env.timeout(self.time_model.get_next_time())

    def interrupt_process(self):
        pass


class SetupState(State):
    """
    Represents a setup state of a resource in the simulation. A setup state has a process that simulates the setup of a resource. This changes the current setup of the resource and allows it processing of other types of processes with their associated production or transport states.

    Args:
        state_data (SetupStateData): The data of the state.
        time_model (time_model.TimeModel): The time model of the state.
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indidcates if the state is active. Defaults to None.
        finished_process (events.Event, optional): Event that indicates if the state is finished. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
        start (float, optional): The start time of the state. Defaults to 0.0.
        done_in (float, optional): The time the state is done in. Defaults to 0.0.

    Attributes:
        interrupt_processed (events.Event): Event that indicates if the state is interrupted. Defaults to None.
    """
    state_data: SetupStateData
    start: float = 0.0
    done_in: float = 0.0
    interrupted: bool = False

    def prepare_for_run(self):
        self.finished_process = events.Event(self.env)

    def activate_state(self):
        self.active = events.Event(self.env).succeed()

    def process_state(self) -> Generator:
        self.done_in = self.time_model.get_next_time()
        while True:
            try:
                if self.interrupted:
                    debug_logging(self, f"interrupted while waiting for activation or activation of resource")
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    self.interrupted = False
                debug_logging(self, f"wait for activation or activation of resource before process")
                yield events.AllOf(self.env, [self.resource.active, self.active])
                running_processes = [
                    state.process
                    for state in self.resource.production_states
                    if (state.process and state.process.is_alive)
                ]
                debug_logging(self, f"waiting for running processes to finish")
                yield events.AllOf(self.env, running_processes)
                break
            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(f"Simpy interrupt occured at {self.state_data.ID} although process is not interrupted")
        while self.done_in:
            try:
                if self.interrupted:
                    self.state_info.log_start_interrupt_state(
                        self.env.now, StateTypeEnum.setup
                    )
                    self.update_done_in()
                    debug_logging(self, f"interrupted process that ends in {self.done_in}")
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    self.interrupted = False
                    debug_logging(self, f"interrupt over for process that ends in {self.done_in}")
                    self.state_info.log_end_interrupt_state(
                        self.env.now, self.env.now + self.done_in, StateTypeEnum.setup
                    )
                self.start = self.env.now
                self.state_info.log_start_state(
                        self.start, self.start + self.done_in, StateTypeEnum.setup
                )
                debug_logging(self, f"starting process that ends in {self.done_in}")
                yield self.env.timeout(self.done_in)
                self.done_in = 0
            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(f"Simpy interrupt occured at {self.state_data.ID} although process is not interrupted")
        debug_logging(self, f"process finished")
        self.state_info.log_end_state(self.env.now, StateTypeEnum.setup)
        self.finished_process.succeed()

    def update_done_in(self):
        self.done_in -= (self.env.now - self.start)  # How much time left?
        if self.done_in < 0:
            self.done_in = 0

    def interrupt_process(self) -> Generator:
        if self.process and self.process.is_alive and not self.interrupted:
            self.interrupted = True
            self.process.interrupt()


MINIMUM_BATTERY_LEVEL = 0.1

class ChargingState(State):
    """
    Represents a charging state of a resource in the simulation. A charging state has a process that simulates the charging of a resource. This changes the current charge of the resource and allows it processing of other types of processes with their associated production or transport

    Args:
        state_data (ChargingStateData): The data of the state.
        time_model (time_model.TimeModel): The time model of the state.
        battery_time_model (time_model.TimeModel): The time model of the battery.
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indidcates if the state is active. Defaults to None.
        finished_process (events.Event, optional): Event that indicates if the state is finished. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
        start (float, optional): The start time of the state. Defaults to 0.0.
        done_in (float, optional): The ramaining time for the state to finish. Defaults to 0.0.
        interrupted (bool, optional): Indicates if the state is interrupted. Defaults to False.
    """
    state_data: ChargingStateData
    battery_time_model: time_model.TimeModel
    start: float = 0.0
    done_in: float = 0.0
    interrupted: bool = False

    battery_usage_time_since_charging: float = 0.0

    def requires_charging(self) -> bool:
        return self.battery_usage_time_since_charging >= (1 - MINIMUM_BATTERY_LEVEL) * self.battery_time_model.get_next_time()

    def prepare_for_run(self):
        self.finished_process = events.Event(self.env)

    def activate_state(self):
        self.active = events.Event(self.env).succeed()

    def add_battery_usage_time(self, time: float):
        self.battery_usage_time_since_charging += time

    def process_state(self) -> Generator:
        self.done_in = self.time_model.get_next_time()
        while True:
            try:
                if self.interrupted:
                    debug_logging(self, f"interrupted while waiting for activation or activation of resource")
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    self.interrupted = False
                debug_logging(self, f"wait for activation or activation of resource before process")
                yield events.AllOf(self.env, [self.resource.active, self.active])
                break
            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(f"Simpy interrupt occured at {self.state_data.ID} although process is not interrupted")
        while self.done_in:
            try:
                if self.interrupted:
                    self.state_info.log_start_interrupt_state(
                        self.env.now, StateTypeEnum.production
                    )
                    self.update_done_in()
                    debug_logging(self, f"interrupted process that ends in {self.done_in}")
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    debug_logging(self, f"interrupt over for process that ends in {self.done_in}")
                    self.interrupted = False
                    self.state_info.log_end_interrupt_state(
                        self.env.now, self.env.now + self.done_in, StateTypeEnum.charging
                    )
                self.start = self.env.now
                self.state_info.log_start_state(
                        self.start, self.start + self.done_in, StateTypeEnum.charging
                )
                debug_logging(self, f"starting process that ends in {self.done_in}")
                yield self.env.timeout(self.done_in)
                self.done_in = 0  # Set to 0 to exit while loop.

            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(f"Simpy interrupt occured at {self.state_data.ID} although process is not interrupted")
        debug_logging(self, f"process finished")
        self.state_info.log_end_state(self.env.now, StateTypeEnum.charging)
        self.battery_usage_time_since_charging = 0
        self.finished_process.succeed()

    def update_done_in(self):
        if self.start == 0:
            return
        self.done_in -= (self.env.now - self.start)  # How much time left?
        if self.done_in < 0:
            self.done_in = 0

    def interrupt_process(self):
        if self.process and self.process.is_alive and not self.interrupted:
            self.interrupted = True
            self.process.interrupt()


STATE_UNION = Union[
    ChargingState, BreakDownState, ProductionState, TransportState, SetupState, ProcessBreakDownState
]
"""
Union Type of all states.
"""

from prodsys.simulation import resources

if TYPE_CHECKING:
    from prodsys.simulation import product


from __future__ import annotations

from typing import List, Optional, Union
from uuid import uuid1

from pydantic import Field
from pydantic.dataclasses import dataclass

from prodsys.models import product_data

from prodsys.express import core, process, auxiliary

@dataclass
class Product(core.ExpressObject):
    """
    Class that represents a product.

    Args:
        processes (List[Union[process.ProductionProcess, process.CapabilityProcess]]): Processes of the product required for its production. These processes are executed sequentially. To have more degree of freedom with precedence graph process models, use the prodsys.datastructures API of product.
        transport_process (process.TransportProcess): Transport process of the product.
        ID (str): ID of the product.

    Attributes:
        processes (List[Union[process.ProductionProcess, process.CapabilityProcess]]): Processes of the product required for its production.
        transport_process (process.TransportProcess): Transport process of the product.
        ID (Optional[str]): ID of the product.
        auxiliaries (Optional[List[auxiliary.Auxiliary]]): Auxiliary processes of the product.


    Examples:
        Product with 2 sequential processes and a transport process:
        ```py
        import prodsys.express as psx
        welding_time_model = psx.time_model_data.FunctionTimeModel(
            distribution_function="normal",
            location=20.0,
            scale=5.0,
         )
        welding_process_1 = psx.process.ProductionProcess(
            time_model=welding_time_model,
        )
        welding_process_2 = psx.process.ProductionProcess(
            time_model=welding_time_model,
        )
        transport_time_model = psx.time_model_data.ManhattenDistanceTimeModel(
            speed=10,
            reaction_time= 0.3
        )
        transport_process = psx.process.TransportProcess(
            time_model=transport_time_model,
        )
        psx.Product(
            processes=[welding_process_1, welding_process_2],
            transport_process=transport_process
        )
        ```

    """

    processes: List[Union[process.ProductionProcess, process.CapabilityProcess, process.RequiredCapabilityProcess]]
    transport_process: Union[process.TransportProcess, process.RequiredCapabilityProcess]
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    auxiliaries: Optional[List[auxiliary.Auxiliary]]= Field(default_factory=list)

    def to_model(self) -> product_data.ProductData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            product_data.ProductData: An instance of the data object.
        """
        return product_data.ProductData(
            ID=self.ID,
            description="",
            processes=[process.ID for process in self.processes],
            transport_process=self.transport_process.ID,
            auxiliaries=[auxiliary.ID for auxiliary in self.auxiliaries]
        )

from prodsys.express import process


from typing import TYPE_CHECKING, List

from pydantic import BaseModel

from prodsys.models.node_data import NodeData


class Node(BaseModel):
    data: NodeData

    def get_location(self) -> List[float]:
        """
        Returns the location of the resource.

        Returns:
            List[float]: The location of the resource. Has to have length 2.
        """
        return self.data.location


from typing import Dict, List, Union
from prodsys import adapters, runner
from prodsys.adapters.adapter import assert_no_redudant_locations, assert_required_processes_in_resources_available, get_possible_production_processes_IDs
from prodsys.models import performance_indicators
from prodsys.optimization.util import get_grouped_processes_of_machine, get_num_of_process_modules, get_required_auxiliaries, get_weights
from prodsys.util.post_processing import PostProcessor


def get_reconfiguration_cost(
    adapter_object: adapters.ProductionSystemAdapter,
    baseline: adapters.ProductionSystemAdapter = None,
) -> float:
    num_machines = len(adapters.get_machines(adapter_object))
    num_transport_resources = len(adapters.get_transport_resources(adapter_object))
    num_process_modules = get_num_of_process_modules(adapter_object)
    if not baseline:
        num_machines_before = 4
        num_transport_resources_before = 1
        possible_processes = get_possible_production_processes_IDs(adapter_object)
        num_process_modules_before = {}
        for process in possible_processes:
            if isinstance(process, str):
                num_process_modules_before[tuple(process)] = 0
            else:
                num_process_modules_before[process] = 0
    else:
        num_machines_before = len(adapters.get_machines(baseline))
        num_transport_resources_before = len(adapters.get_transport_resources(baseline))
        num_process_modules_before = get_num_of_process_modules(baseline)

    if adapter_object.auxiliary_data:
        auxiliary_cost = get_auxiliary_cost(adapter_object, baseline)
    else:
        auxiliary_cost = 0

    machine_cost = max(
        0,
        (num_machines - num_machines_before)
        * adapter_object.scenario_data.info.machine_cost,
    )
    transport_resource_cost = max(
        0,
        (num_transport_resources - num_transport_resources_before)
        * adapter_object.scenario_data.info.transport_resource_cost,
    )
    process_module_cost = 0
    for process in num_process_modules:
        process_module_cost += max(
            0,
            (num_process_modules[process] - num_process_modules_before[process])
            * adapter_object.scenario_data.info.process_module_cost,
        )

    return machine_cost + transport_resource_cost + process_module_cost + auxiliary_cost


def get_auxiliary_cost(
    adapter_object: adapters.ProductionSystemAdapter,
    baseline: adapters.ProductionSystemAdapter,
) -> float:
    auxiliary_cost = 0
    for new_auxiliary, auxiliary_before in zip(adapter_object.auxiliary_data, baseline.auxiliary_data):
        for i, storage in enumerate(new_auxiliary.quantity_in_storages):
            storage_before = auxiliary_before.quantity_in_storages[i]
            auxiliary_cost += max(
                0,
                (storage - storage_before)
                * adapter_object.scenario_data.info.auxiliary_cost,
            )
    return auxiliary_cost


def valid_num_machines(configuration: adapters.ProductionSystemAdapter) -> bool:
    if (
        len(adapters.get_machines(configuration))
        > configuration.scenario_data.constraints.max_num_machines
    ):
        return False
    return True


def valid_transport_capacity(configuration: adapters.ProductionSystemAdapter) -> bool:
    if (
        len(adapters.get_transport_resources(configuration))
        > configuration.scenario_data.constraints.max_num_transport_resources
    ) or (len(adapters.get_transport_resources(configuration)) == 0):
        return False
    return True


def valid_num_process_modules(configuration: adapters.ProductionSystemAdapter) -> bool:
    for resource in configuration.resource_data:
        if (
            len(
                get_grouped_processes_of_machine(
                    resource, get_possible_production_processes_IDs(configuration)
                )
            )
            > configuration.scenario_data.constraints.max_num_processes_per_machine
        ):
            return False
    return True


def valid_positions(configuration: adapters.ProductionSystemAdapter) -> bool:
    try:
        assert_no_redudant_locations(configuration)
    except ValueError as e:
        return False

    positions = [machine.location for machine in adapters.get_machines(configuration)]
    possible_positions = configuration.scenario_data.options.positions
    if any(position not in possible_positions for position in positions):
        return False
    return True


def valid_reconfiguration_cost(
    configuration: adapters.ProductionSystemAdapter,
    base_configuration: adapters.ProductionSystemAdapter,
) -> bool:
    reconfiguration_cost = get_reconfiguration_cost(
        adapter_object=configuration,
        baseline=base_configuration,
    )
    configuration.reconfiguration_cost = reconfiguration_cost

    if (
        reconfiguration_cost
        > configuration.scenario_data.constraints.max_reconfiguration_cost
    ):
        return False
    return True


def check_valid_configuration(
    configuration: adapters.ProductionSystemAdapter,
    base_configuration: adapters.ProductionSystemAdapter,
) -> bool:
    """
    Function that checks if a configuration is valid.

    Args:
        configuration (adapters.ProductionSystemAdapter): Configuration to be checked.
        base_configuration (adapters.ProductionSystemAdapter): Baseline configuration.

    Returns:
        bool: True if the configuration is valid, False otherwise.
    """
    if not valid_num_machines(configuration):
        return False
    if not valid_transport_capacity(configuration):
        return False
    if not valid_num_process_modules(configuration):
        return False
    try:
        assert_required_auxiliaries_available(configuration)
    except ValueError as e:
        return False
    try:
        assert_required_processes_in_resources_available(configuration)
    except ValueError as e:
        return False
    if not valid_positions(configuration):
        # TODO: raise error if the positions cannot be changed (no production capacity or layout in transformations of scenario)
        return False
    if not valid_reconfiguration_cost(configuration, base_configuration):
        return False
    return True


def assert_required_auxiliaries_available(
    configuration: adapters.ProductionSystemAdapter,
) -> bool:
    required_auxiliaries = get_required_auxiliaries(configuration)
    for auxiliary in required_auxiliaries:
        if not sum(auxiliary.quantity_in_storages) > 0:
            raise ValueError(f"Required auxiliary {auxiliary.ID} is not available.")
        


def get_throughput_time(pp: PostProcessor) -> float:
    throughput_time_for_products = pp.get_aggregated_throughput_time_data()
    if not throughput_time_for_products:
        throughput_time_for_products = [100000]
    avg_throughput_time = sum(throughput_time_for_products) / len(
        throughput_time_for_products
    )
    return avg_throughput_time


def get_wip(pp: PostProcessor) -> float:
    return sum(pp.get_aggregated_wip_data())


def get_throughput(pp: PostProcessor) -> float:
    return sum(pp.get_aggregated_throughput_data())


KPI_function_dict = {
    performance_indicators.KPIEnum.COST: get_reconfiguration_cost,
    performance_indicators.KPIEnum.TRHOUGHPUT_TIME: get_throughput_time,
    performance_indicators.KPIEnum.WIP: get_wip,
    performance_indicators.KPIEnum.THROUGHPUT: get_throughput,
}


def evaluate(
    base_scenario: adapters.ProductionSystemAdapter,
    solution_dict: Dict[str, Union[list, str]],
    performances: dict,
    number_of_seeds: int,
    full_save_folder_file_path: str,
    individual,
) -> List[float]:
    """
    Function that evaluates a configuration.

    Args:
        base_scenario (adapters.ProductionSystemAdapter): Baseline configuration.
        solution_dict (Dict[str, Union[list, str]]): Dictionary containing the ids of existing solutions.
        performances (dict): Dictionary containing the performances of the current and previous generations.
        number_of_seeds (int): Number of seeds for the simulation runs.
        individual (List[adapters.ProductionSystemAdapter]): List if length 1 containing the configuration to be evaluated.

    Raises:
        ValueError: If the time range is not defined in the scenario data.

    Returns:
        List[float]: List of the fitness values of the configuration.
    """

    adapter_object: adapters.ProductionSystemAdapter = individual[0]
    adapter_object_hash = adapter_object.hash()
    if adapter_object_hash in solution_dict["hashes"]:
        evaluated_adapter_generation = solution_dict["hashes"][adapter_object_hash]["generation"]
        evaluated_adapter_id = solution_dict["hashes"][adapter_object_hash]["ID"]
        return performances[evaluated_adapter_generation][evaluated_adapter_id]["fitness"]

    if not check_valid_configuration(adapter_object, base_scenario):
        return [-100000 / weight for weight in get_weights(base_scenario, "max")]

    fitness_values = []

    for seed in range(number_of_seeds):
        runner_object = runner.Runner(adapter=adapter_object)
        if not adapter_object.scenario_data.info.time_range:
            raise ValueError("time_range is not defined in scenario_data")
        adapter_object.seed = seed
        runner_object.initialize_simulation()
        runner_object.run(adapter_object.scenario_data.info.time_range)
        if full_save_folder_file_path:
            runner_object.save_results_as_csv(full_save_folder_file_path)
        df = runner_object.event_logger.get_data_as_dataframe()
        p = PostProcessor(df_raw=df)
        fitness = []
        for objective in adapter_object.scenario_data.objectives:
            if objective.name == performance_indicators.KPIEnum.COST:
                fitness.append(get_reconfiguration_cost(adapter_object, base_scenario))
                continue
            fitness.append(KPI_function_dict[objective.name](p))
        fitness_values.append(fitness)

    mean_fitness = [sum(fitness) / len(fitness) for fitness in zip(*fitness_values)]
    return mean_fitness


"""
`process` module contains the `prodsys.express` classes to represent the processes that can 
be performed on products by resources.

The following processes are possible:
- `ProductionProcess`: A process that can be performed on a product by a production resource.
- `CapabilityProcess`: A process that can be performed on a product by a resource, based on the capability of the resource.
- `TransportProcess`: A process that can be performed on a product by a transport resource.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, Union
from uuid import uuid1

from abc import ABC

from pydantic import Field
from pydantic.dataclasses import dataclass

from prodsys.express import core, time_model
from prodsys.models import processes_data

@dataclass
class Process(ABC):
    """
    Abstract base class to represents a process.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the process.

    Attributes:
        type (processes_data.ProcessTypeEnum): Type of the process.
    """

    time_model: time_model.TIME_MODEL_UNION

@dataclass
class DefaultProcess(Process):
    """
    Abstract base class to represents a process, with no additional attributes than type and ID.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the process.
        ID (str): ID of the process.
    
    Attributes:
        type (processes_data.ProcessTypeEnum): Type of the process.
    """
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    type: processes_data.ProcessTypeEnum = Field(init=False)




@dataclass
class ProductionProcess(DefaultProcess, core.ExpressObject):
    """
    Class that represents a production process.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the process.
        ID (str): ID of the process.

    Attributes:
        type (processes_data.ProcessTypeEnum): Type of the process. Equals to processes_data.ProcessTypeEnum.ProductionProcesses.

    Examples:
        Production process with a function time model:
        ``` py
        import prodsys.express as psx
        welding_time_model = psx.time_model_data.FunctionTimeModel(
            distribution_function="normal",
            location=20.0,
            scale=5.0,
        )
        psx.ProductionProcess(
            time_model=welding_time_model,
        )
        ```
    """
    type: processes_data.ProcessTypeEnum = Field(
        init=False, default=processes_data.ProcessTypeEnum.ProductionProcesses
    )

    def to_model(self) -> processes_data.ProductionProcessData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            processes_data.ProductionProcessData: Data object of the express object.
        """
        return processes_data.ProductionProcessData(
            time_model_id=self.time_model.ID,
            ID=self.ID,
            description="",
            type=self.type
        )


@dataclass
class CapabilityProcess(Process, core.ExpressObject):
    """
    Class that represents a capability process. For capability processes, matching of 
    required processes of product and provided processes by resources is done based on 
    the capability instead of the porcess itself.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the process.
        capability (str): Capability of the process.
        ID (str): ID of the process.

    Attributes:
        type (processes_data.ProcessTypeEnum): Type of the process. Equals to processes_data.ProcessTypeEnum.CapabilityProcesses.

    Examples:
        Capability process with a function time model:
        ``` py
        import prodsys.express as psx
        welding_time_model = psx.FunctionTimeModel(
            distribution_function="normal",
            location=20.0,
            scale=5.0,
        )
        psx.CapabilityProcess(
            time_model=welding_time_model,
            capability="welding"
        )
        ```
    """
    capability: str
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    type: processes_data.ProcessTypeEnum = Field(
        init=False, default=processes_data.ProcessTypeEnum.CapabilityProcesses
    )

    def to_model(self) -> processes_data.CapabilityProcessData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            processes_data.CapabilityProcessData: Data object of the express object.
        """
        return processes_data.CapabilityProcessData(
            time_model_id=self.time_model.ID,
            capability=self.capability,
            ID=self.ID,
            description="",
            type=self.type
        )

@dataclass
class TransportProcess(DefaultProcess, core.ExpressObject):
    """
    Class that represents a transport process. Transport processes are required to transport product from one location to another. 

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the process.
        ID (str): ID of the process.

    Attributes:
        type (processes_data.ProcessTypeEnum): Type of the process. Equals to processes_data.ProcessTypeEnum.TransportProcesses.

    Examples:
        Transport process with a manhattan distance time model:
        ```py
        import prodsys.express as psx
        manhattan_time_model = psx.ManhattenDistanceTimeModel(
            speed=30.0,
            reaction_time=0.15,
        )
        psx.TransportProcess(
            time_model=manhattan_time_model
        )
        ```
    """ 
    type: processes_data.ProcessTypeEnum = Field(
        init=False, default=processes_data.ProcessTypeEnum.TransportProcesses
    )

    def to_model(self) -> processes_data.TransportProcessData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            processes_data.TransportProcessData: Data object of the express object.
        """
        return processes_data.TransportProcessData(
            time_model_id=self.time_model.ID,
            ID=self.ID,
            description="",
            type=self.type
        )


@dataclass
class LinkTransportProcess(TransportProcess):
    """
    Represents a link transport process. They include a list or dict of links.

    Attributes:
        type (processes_data.ProcessTypeEnum): The type of the process.
        ID (Optional[str]): The ID of the process.
        links (Union[List[List[Union[resources.Resource, resources.NodeData, source.Source, sink.Sink]]], 
                      Dict[Union[resources.Resource, resources.NodeData, source.Source, sink.Sink], 
                           List[Union[resources.Resource, resources.NodeData, source.Source, sink.Sink]]]]):
            The links associated with the process. Each link is a list of different objects, which can be a 
            Resource, NodeData, Source, or Sink. If the links attribute is a list, it represents a list of 
            links, where each link is a list of these objects. If the links attribute is a dictionary, it represents a 
            mapping from a key (which can be a ProductionResource, NodeData, Source, or Sink) to a list of these objects.
        capability (Optional[str]): The capability of the process.
    """

    type: processes_data.ProcessTypeEnum = Field(
        init=False, default=processes_data.ProcessTypeEnum.LinkTransportProcesses
    )
    links: Union[List[List[Union[resources.Resource, sink.Sink, source.Source, node.Node]]], 
                 Dict[Union[resources.Resource, sink.Sink, source.Source, node.Node], 
                      List[Union[resources.Resource, sink.Sink, source.Source, node.Node]]]] = Field(default_factory=list)
    capability: Optional[str] = Field(default_factory=str)

    def add_link(self, link: List[Union[resources.Resource, sink.Sink, source.Source, node.Node]]) -> None:
        """
        Adds a link to the LinkTransportProcess object.

        Args:
            link (Union[Resource, Node, Source, Sink]): The link to add.
            link_list (List[Union[Resource, Node, Source, Sink]]): The list of links to add the link to.
        """
        if isinstance(self.links, list):
            self.links.append(link)
        else:
            if not link in self.links:
                self.links[link] = []
            self.links[link[0]] = link[1]

    def set_links(self, links: List[List[Union[resources.Resource, node.Node, source.Source, sink.Sink]]]) -> None:
        """
        Sets the links of the LinkTransportProcess object.

        Args:
            links (List[List[Union[Resource, Node, Source, Sink]]]): The links to set.
        """
        self.links = links

    def to_model(self) -> processes_data.LinkTransportProcessData:
        """
        Converts the LinkTransportProcess object to its corresponding data model.

        Returns:
            processes_data.LinkTransportProcessData: The converted data model object.
        """
        if isinstance(self.links, list):
            return_links = [[link.ID for link in link_list] for link_list in self.links]
        else:
            return_links = []
            for start, targets in self.links.items():
                for target in targets:
                    return_links.append([start.ID, target.ID])

        return processes_data.LinkTransportProcessData(
            time_model_id=self.time_model.ID,
            ID=self.ID,
            description="",
            type=self.type,
            links=return_links,
            capability=self.capability,
        )


@dataclass
class RequiredCapabilityProcess(core.ExpressObject):
    """
    Represents a required capability process. A capability which can be matched with the capability of another process with a capability.

    Attributes:
        ID (Optional[str]): The ID of the process.
        type (processes_data.ProcessTypeEnum): The type of the process.
        capability (Optional[str]): The capability required by the process.
    """
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    type: processes_data.ProcessTypeEnum = Field(
        init=False, default=processes_data.ProcessTypeEnum.RequiredCapabilityProcesses
    )
    capability: Optional[str] = Field(default_factory=str)

    def to_model(self) -> processes_data.RequiredCapabilityProcessData:
        """
        Converts the RequiredCapabilityProcess object to its corresponding data model.

        Returns:
            processes_data.RequiredCapabilityProcessData: The data model representation of the process.
        """
        return processes_data.RequiredCapabilityProcessData(
            ID=self.ID,
            description="",
            type=self.type,
            capability=self.capability,
        )



PROCESS_UNION = Union[
    ProductionProcess,
    CapabilityProcess,
    TransportProcess,
    RequiredCapabilityProcess,
    LinkTransportProcess,
]
from prodsys.express import resources, sink, source, node



from __future__ import annotations

from abc import ABC, abstractmethod

from functools import partial
from typing import TYPE_CHECKING, List, Optional, Tuple, Callable, Union

import numpy as np
from simpy import events

import gymnasium as gym
from gymnasium import spaces
from prodsys import adapters, runner
from prodsys.simulation import control, sim, observer, router

sim.VERBOSE = 0

if TYPE_CHECKING:
    from prodsys.simulation import resources, request


class AbstractSequencingControlEnv(gym.Env, ABC):
    """
    Abstract Gym environment for controlling a controller of a resource wtih an agent.
    This class defines the methods that need to be implemented in order to use a Reinforcement learning agent for production control.

    Args:
        adapter (adapters.ProductionSystemAdapter): The adapter.
        resource_id (str): The ID of the resource to control.
        observation_space (Optional[spaces.Space], optional): The observation space of the environment.
        action_space (Optional[spaces.Space], optional): The action space of the environment.
        render_mode (Optional[str], optional): The render mode of the environment. Defaults to None.

    Attributes:
        adapter (adapters.ProductionSystemAdapter): The adapter.
        resource_id (str): The ID of the resource to control.
        observation_space (Optional[spaces.Space], optional): The observation space of the environment.
        action_space (Optional[spaces.Space], optional): The action space of the environment.
        render_mode (Optional[str], optional): The render mode of the environment. Defaults to None.
        runner (runner.Runner): The runner of the adapter.
        interrupt_simulation_event (events.Event): The event to interrupt the simulation when an agent interaction is needed.
        resource_controller (control.Controller): The controller of the resource.
        resource (resources.Resource): The resource to control.
        observer (observer.ResourceObserver): The observer of the resource.
        step_count (int): The number of steps taken in the environment.
        reward (float): The reward of the environment.
    """

    def __init__(
        self,
        adapter: adapters.ProductionSystemAdapter,
        resource_id: str,
        observation_space: Optional[spaces.Space] = None,
        action_space: Optional[spaces.Space] = None,
        render_mode: Optional[str]=None,
    ):
        self.adapter = adapter
        self.resource_id = resource_id
        self.observation_space = observation_space
        self.action_space = action_space
        self.render_mode = render_mode

        self.runner = runner.Runner(adapter=self.adapter)

        self.interrupt_simulation_event: events.Event = None
        self.resource_controller: control.Controller = None
        self.resource: Union[resources.ProductionResource, resources.TransportResource] = None
        self.observer: observer.ResourceObserver = None
        self.step_count: int = 0
        self.reward = 0

    @abstractmethod
    def get_observation(self) -> np.ndarray:
        """
        Get observation of the environment.

        Returns:
            np.ndarray: The observation.
        """
        pass

    @abstractmethod
    def get_info(self) -> dict:
        """
        Get info of the environment.

        Returns:
            dict: The info.
        """
        pass

    @abstractmethod
    def get_reward(self, processed_request: request.Request, invalid_action:bool=False) -> float:
        """
        Get reward of the environment.

        Args:
            processed_request (request.Request): The processed request of the last step.
            invalid_action (bool, optional): Whether the last action was invalid. Defaults to False.

        Returns:
            float: The reward.
        """
        pass

    @abstractmethod
    def get_termination_condition(self) -> bool:
        """
        Get termination condition of the environment.

        Returns:
            bool: The termination condition.
        """
        pass

    def reset(self, seed=None):
        """
        Reset env for new episode and run until first point of observation.

        Args:
            seed (Optional[int], optional): The seed for the environment. Defaults to None.
            options (Optional[dict], optional): The options for the environment. Defaults to None.
        """
        super().reset(seed=seed)

        self.runner.initialize_simulation()
        self.interrupt_simulation_event = events.Event(self.runner.env)
        self.resource_controller = self.runner.resource_factory.get_resource(
            self.resource_id
        ).get_controller()
        self.resource = self.resource_controller.resource
        control_policy = partial(control.agent_control_policy, self)
        self.resource_controller.control_policy = control_policy
        self.observer = observer.ResourceObserver(
            resource_factory=self.runner.resource_factory,
            product_factory=self.runner.product_factory,
            resource=self.resource,
        )

        self.runner.env.run_until(until=self.interrupt_simulation_event)
        self.interrupt_simulation_event = events.Event(self.runner.env)

        observation = self.get_observation()
        info = self.get_info()

        if self.render_mode == "human":
            self.render()

        return observation, info

    def step(self, action: np.ndarray) -> Tuple[np.ndarray, float, bool, dict]:
        """
        Take a step in the environment.

        Args:
            action (np.ndarray): The output of the agent for actions.

        Returns:
            Tuple[np.ndarray, float, bool, dict]: The observation, reward, done, and info.
        """
        queue_index = np.argmax(action)

        if queue_index >= len(self.resource_controller.requests):
            queue_index = np.random.choice(
                [i for i in range(len(self.resource_controller.requests))]
            )
            to_process = self.resource_controller.requests.pop(queue_index)
            invalid_action = True
        else:
            to_process = self.resource_controller.requests.pop(queue_index)
            invalid_action = False

        self.resource_controller.requests.insert(0, to_process)
        self.runner.env.run_until(until=self.interrupt_simulation_event)
        self.step_count += 1
        self.interrupt_simulation_event = events.Event(self.runner.env)

        terminated = self.get_termination_condition()
        self.reward = self.get_reward(to_process, invalid_action)
        observation = self.get_observation()
        info = self.get_info()

        if self.render_mode == "human":
            self.render()

        return observation, self.reward, terminated, False, info
    
    def render(self):
        """
        Render the environment.
        """
        if self.render_mode == "human":
            pass
            
    def set_observation_space(self, observation_space: spaces.Space):
        """
        Set the observation space of the environment.

        Args:
            observation_space (spaces.Space): The observation space of the environment.
        """
        self.observation_space = observation_space

    def set_action_space(self, action_space: spaces.Space):
        """
        Set the action space of the environment.

        Args:
            action_space (spaces.Space): The action space of the environment.
        """
        self.action_space = action_space

    def set_binary_box_observation_space_from_shape(self, shape: Tuple[int, ...]):
        """
        Set the observation space of the environment to a binary box space.

        Args:
            shape (Tuple[int, ...]): The shape of the observation space.
        """
        self.observation_space = spaces.Box(0, 1, shape=shape, dtype=int)


    def set_binary_box_action_space_from_shape(self, shape: Tuple[int, ...]):
        """
        Set the action space of the environment to a binary box space.

        Args:
            shape (Tuple[int, ...]): The shape of the action space.
        """
        self.action_space = spaces.Box(0, 1, shape=shape, dtype=float)



from __future__ import annotations

from typing import Dict, List, TYPE_CHECKING, Optional, Tuple, Union

import numpy as np

from pathfinding.core.graph import Graph, GraphNode
from pathfinding.finder.dijkstra import DijkstraFinder

if TYPE_CHECKING:
    from prodsys.simulation import request, process, resources
    from prodsys.simulation.product import Locatable


def find_route(request: request.TransportResquest, process: process.LinkTransportProcess, find_route_to_origin: bool=False) -> List[Locatable]:
    """
    Finds the route for a transportation request.

    Args:
        request (request.TransportResquest): The transportation request.
        process (process.LinkTransportProcess): The process to find the route for.
        find_route_to_origin (bool, optional): Indicates whether to find the route from current resource location to origin (True) or from origin to target of request (False). Defaults to False.

    Returns:
        List[Locatable]: The route as a list of locatable objects.
    """
    if request.route and not find_route_to_origin:
        return request.route
    route_finder = RouteFinder()
    route = route_finder.find_route(request=request, process=process, find_route_to_origin=find_route_to_origin)
    if route and not find_route_to_origin:
        request.set_route(route=route)
    return route

class RouteFinder:
    """
    Class representing a route finder for transportation requests in a graph.
    """

    def __init__(self):
        """
        Initializes two empty lists of nodes and node locations.
        """
        self.nodes: Dict[str, GraphNode] = {}
        self.node_locatables: Dict[Tuple[float,float], Locatable] = {}

    def find_route(self, request: request.TransportResquest, process: process.LinkTransportProcess, find_route_to_origin: bool=False) -> List[Locatable]:
        """
        The general function which includes all sub functions to find the shortest route for a TransportRequest.

        Args:
            request (TransportResquest): The transportation request.
            process (LinkTransportProcess): The process to find the route for.
            find_route_to_origin (bool): Indicates whether to find the route from current resource location to origin (True) or from origin to target of request (False).

        Returns:
            List[Locatable]: The route as a list of locatable objects.
        """
        edges = self.process_links_to_graph_edges(links=process.links)
        # TODO: also add functionality to make directional edges for conveyors, where backwards routing is not possibly. 
        # For this feature is it necessary, that the location of the conveyor does not change after a transport.
        graph = Graph(edges=edges, bi_directional=True)
        origin, target = self.get_route_origin_and_target(request=request, route_to_origin=find_route_to_origin)
        if not origin or not target:
            return []
        graph_node_path = self.find_graphnode_path(origin, target, graph)
        if not graph_node_path:
            return []
        route = self.convert_node_path_to_locatable_route(graph_node_path=graph_node_path, links=process.links)

        return route

    def process_links_to_graph_edges(self, links: List[List[Locatable]]) -> List[Tuple[GraphNode, GraphNode, int]]:
        """
        Processes the given links to create (Graph)-edges for the graph.

        Args:
            links (List[List[Locatable]]): The links as a list of lists of locatable objects.

        Returns:
            List[Tuple[Graphnode, Graphnode, int]]: The edges as a list of tuples (with a start_node, end_node and related costs).
        """
        pathfinder_edges = []

        for link in links:
            cost = self.calculate_cost(link[0].get_location(), link[1].get_location())
            origin_graph_node, target_graph_node = self.get_graph_nodes_for_link(link)
            edge = (origin_graph_node, target_graph_node, cost)
            pathfinder_edges.append(edge)

        return pathfinder_edges
    
    def get_graph_nodes_for_link(self, link: List[Locatable]) -> Tuple[GraphNode, GraphNode]:
        """
        Creates a list of links and edges.

        Args:
            link (List[Locatable]): The link as a list of two locatable objects.

        Returns:
            Tuple[GraphNode, GraphNode]: A tuple containing the origin and target nodes for a link.
        """
        graph_nodes = []
        for locatable in link:
            graph_node = self.get_graph_node_for_locatable(locatable)
            graph_nodes.append(graph_node)
        return tuple(graph_nodes)
    

    def get_existing_graph_node_for_locatable(self, locatable: Locatable) -> Optional[GraphNode]:
        """
        Gets an existing graph node for a Locatable.

        Args:
            locatable (Locatable): The locatable object to get the graph node for.

        Returns:
            Optional[GraphNode, None]: The graph node or None.
        """
        if locatable.data.ID in self.nodes:
            return self.nodes[locatable.data.ID]
        return None
    
    def get_graph_node_for_locatable(self, locatable: Locatable) -> GraphNode:
        """
        Creates a graph node for a locatable.

        Args:
            locatable (Locatable): The locatable to create the graph node for.

        Returns:
            GraphNode: The graph node.
        """
        existing_node = self.get_existing_graph_node_for_locatable(locatable)
        if existing_node:
            return existing_node
        new_graph_node = GraphNode(node_id=locatable.data.ID)
        self.nodes[locatable.data.ID] = new_graph_node
        self.node_locatables[tuple(locatable.get_location())] = locatable
        return new_graph_node
        

    def calculate_cost(self, node1: List[float], node2: List[float]) -> float:
        """
        Calculates the cost between two nodes based on the euclidean distance.

        Args:
            node1 (List[float]): The first node.
            node2 (List[float]): The second node.

        Returns:
            float: The cost.
        """
        # TODO: maybe use here the time model feature to weight the distance with the time it takes to travel it to optimize for fastest paths and not shortest paths
        return np.sqrt((node1[0] - node2[0])**2 + (node1[1] - node2[1])**2)

    def get_route_origin_and_target(self, request: request.TransportResquest, route_to_origin: bool) -> Tuple[Optional[GraphNode], Optional[GraphNode]]:
        """
        Converts the origin and target of the transport request to graph nodes.

        Args:
            request (TransportResquest): The transportation request.
            route_to_origin (bool): Indicates whether to find the route from current resource location to origin (True) or from origin to target of request (False).

        Returns:
            Tuple[Optional[GraphNode], Optional[GraphNode]]: A tuple containing the origin and target graph nodes for the transport request and the route_to_origin flag.
        """
        if route_to_origin:
            origin_locatable = self.get_location_of_transport_resource(request.resource)
            target_locatable = request.origin
        else:
            origin_locatable = request.origin
            target_locatable = request.target
        origin_graph_node = self.get_existing_graph_node_for_locatable(origin_locatable)
        target_graph_node = self.get_existing_graph_node_for_locatable(target_locatable)
        return origin_graph_node, target_graph_node
    
    def get_location_of_transport_resource(self, resource: resources.Resource) -> Locatable:
        """
        Gets the location of a resource.

        Args:
            resource (Resource): The resource.

        Returns:
            Locatable: The locatable object where the transport resource currently is.
        """
        resource_location = tuple(resource.get_location())
        if resource_location in self.node_locatables:
            return self.node_locatables[resource_location]
        raise ValueError(f"The current location for resource {resource.data.ID} could not be found. {resource.data.ID} is at location {resource.get_location()} that is not part of the links.")

    def find_graphnode_path(self, origin: GraphNode, target: GraphNode, graph: Graph) -> List[GraphNode]:
        """
        Finds the path between the origin and target graph nodes.

        Args:
            origin (GraphNode): The origin node.
            target (GraphNode): The target node.
            graph (Graph): The graph.

        Returns:
            The path as a list of graph nodes.
        """
        finder = DijkstraFinder()
        path, _ = finder.find_path(origin, target, graph)
        if path is None:
            return []
        return path

    def convert_node_path_to_locatable_route(self, graph_node_path: List[GraphNode], links: List[List[Locatable]]) -> List[Locatable]:
        """
        Converts the path of graph nodes to a route of Locatable objects.

        Args:
            graph_node_path (List[GraphNode]): The path as a list of graph nodes.
            links (List[List[Locatable]]): The given links list.

        Returns:
            List[Locatable]: The route as a list of locatable.
        """
        route = []
        seen_node_ids = []  # no node several times in the path

        for node in graph_node_path:
            for link in links:
                for locatable in link:
                    if node.node_id == locatable.data.ID and locatable.data.ID not in seen_node_ids:
                        route.append(locatable)
                        seen_node_ids.append(locatable.data.ID)
        return route
    


"""
NodeData objects are used in prodsys to represent Locations in the production system that serve as nodes in a link. They are used in the 
`LinkTransportProcess` class to represent locations where routes can cross.
"""
from hashlib import md5
from typing import List

from pydantic import ConfigDict, conlist

from prodsys.models.core_asset import CoreAsset

class NodeData(CoreAsset):
    """
    Represents a node data object of a link.

    Attributes:
            location (List[float]): Location of the node. It has to be a list of length 2.
    """

    location: conlist(float, min_length=2, max_length=2) # type: ignore

    def hash(self) -> str:
        """
        Hashes the node data object.

        Returns:
            str: Hash of the node data object.
        """
        return md5("".join([*map(str, self.location)]).encode("utf-8")).hexdigest()
    
    model_config = ConfigDict(json_schema_extra= {
        "examples": [
            {
                "ID": "N1",
                "description": "Node 1",
                "location": [0.0, 0.0],
            }
        ]
    })




import logging
import sys
import logging.config
from datetime import datetime
from typing import Literal, Optional
import os

LOG_FILE_PATH = f'logs/prodsys_debug_{datetime.now().strftime("%Y%m%d_%H%M%S")}.log'
CONFIG_LOCATION = os.path.join(os.path.dirname(__file__), "logging.ini")

logging_handler_dict = {
    "console": "consoleHandler",
    "file": "fileHandler",
    "null": "nullHandler",
}
import logging.handlers

class DelayedRotatingFileHandler(logging.handlers.RotatingFileHandler):
    def __init__(self, filename, mode='a', maxBytes=0, backupCount=0, encoding=None, delay=True):
        super().__init__(filename, mode, maxBytes, backupCount, encoding, delay)

def set_logging(log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "WARNING",
    logging_handler: Literal["null", "console", "file"] = "console",
    Log_file_path: str = LOG_FILE_PATH
):
    logging.config.fileConfig(
        CONFIG_LOCATION,
        defaults={
            "logfilename": Log_file_path,
            "loglevel": log_level,
            "logginghandler": logging_handler_dict[logging_handler],
        },
    )



from abc import ABC, abstractmethod

from prodsys.models import core_asset

class ExpressObject(ABC):
    """
    Abstract base class to represents an express object.
    """

    @abstractmethod
    def to_model(self) -> core_asset.CoreAsset:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            core_asset.CoreAsset: An instance of the data object.
        """
        pass



"""
The `prodsys.models` package contains the data structures to describe a production system's structure and its performance. These data structures are utilized in prodsys to have a common format for all algorithms in simulation, optimization, analysis and the prodsys webserver.
All of these formats are compatible with the `prodsys.express` API. However, the conversion works only in one direction: from `prodsys.express` to `prodsys.models`. 

The following modules are available:

- `prodsys.models.core_asset`: Contains the abstract base class for data objects.
- `prodsys.models.product_data`: Contains classes to represent products.
- `prodsys.models.performance_data`: Contains classes to represent performance data.
- `prodsys.models.performance_indicators`: Contains classes to represent performance indicators (KPIs).
- `prodsys.models.processes_data`: Contains classes to represent processes.
- `prodsys.models.queue_data`: Contains classes to represent queues.
- `prodsys.models.resource_data`: Contains classes to represent resources.
- `prodsys.models.node_data`: Contains classes to represent nodes in a link.
- `prodsys.models.scenario_data`: Contains classes to represent scenario data.
- `prodsys.models.sink_data`: Contains classes to represent sinks.
- `prodsys.models.source_data`: Contains classes to represent sources.
- `prodsys.models.state_data`: Contains classes to represent states.
- `prodsys.models.links_data`: Contains classes to represent links.
- `prodsys.models.time_model_data`: Contains classes to represent time models.
"""